#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024 Pytroll community

# Author(s):

#   Adam Dybbroe <Firstname.Lastname @ smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Create shapefiles of satellite instrument swath outlines from an XML pass reception plan."""


import argparse
import logging
from datetime import datetime
from pathlib import Path

import defusedxml.ElementTree as ET
import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
from shapely.geometry import Polygon
from shapely.ops import transform

from trollsched import INSTRUMENT, SATELLITE_NAMES
from trollsched.logger import setup_logging
from trollsched.satpass import create_pass

logger = logging.getLogger(__name__)


def get_shapely_polygon_from_lonlat(lons, lats):
    """From a arrays of lons and lats return a shapely Polygon."""
    geodata = np.vstack((lons, lats)).T
    return Polygon(np.rad2deg(geodata))


def create_shapefile_filename(satellite_pass_obj):
    """From a trollsched.satpass instance create the shapefile filename."""
    satname = satellite_pass_obj.satellite.name.replace(" ", "-")
    prefix = f"{satellite_pass_obj.instrument}_{satname}"
    return f"{prefix}_{satellite_pass_obj.risetime:%Y%m%d%H%M}_{satellite_pass_obj.falltime:%Y%m%d%H%M}_outline.shp"


def create_shapefile_from_pass(sat_pass, output_filepath):
    """From a satellite overpass (instrument scanning outline) create a shapefile and save."""
    sat_poly = get_shapely_polygon_from_lonlat(sat_pass.boundary.contour_poly.lon,
                                               sat_pass.boundary.contour_poly.lat)

    wgs84 = pyproj.CRS("EPSG:4326")  # WGS 84
    a_crs = {"init": "epsg:4326"}

    project = pyproj.Transformer.from_crs(wgs84, a_crs, always_xy=True).transform
    new_shapes = transform(project, sat_poly)

    gpd.GeoDataFrame(pd.DataFrame(["p1"], columns=["geom"]), crs=a_crs,
                     geometry=[new_shapes]).to_file(output_filepath)


def shapefiles_from_schedule_xml_requests(filename, satellites, tle_file, output_dir):
    """From XML file with planned satellite passes create shapefiles for each Pass outline.

    Open the XML file with the scheduled satellite pass plan, and create
    shapefiles for each Pass outline assuming the instrument as provided in the
    INSTRUMENTS dictionary.
    """
    tree = ET.parse(filename)
    root = tree.getroot()

    for child in root:
        if child.tag == "pass":
            platform_name = SATELLITE_NAMES.get(child.attrib["satellite"], child.attrib["satellite"])
            instrument = INSTRUMENT.get(platform_name)
            if not instrument:
                logger.debug("Instrument not defined for platform %s", platform_name)
                continue

            if platform_name not in satellites:
                logger.debug("Satellite platform name not requested: %s", platform_name)
                continue

            overpass = create_pass(platform_name, instrument,
                                   datetime.strptime(child.attrib["start-time"],
                                                     "%Y-%m-%d-%H:%M:%S"),
                                   datetime.strptime(child.attrib["end-time"],
                                                     "%Y-%m-%d-%H:%M:%S"),
                                   tle_filename=tle_file)

            output_filepath = Path(output_dir) / create_shapefile_filename(overpass)
            create_shapefile_from_pass(overpass, output_filepath)


def parse_args(args):
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-config",
                        dest="log_config",
                        default=None,
                        required=False,
                        help="Log config file to use instead of the standard logging.")
    parser.add_argument("-s", "--satellites",
                        dest="list_of_satellites",
                        nargs="*",
                        default=[],
                        help="Complete file path to the TLE file to use.",
                        required=True)
    parser.add_argument("-t", "--tle_filename",
                        dest="tle_filepath",
                        help="Complete file path to the TLE file to use.",
                        required=True)
    parser.add_argument("-x", "--xml_filename",
                        dest="xml_filepath",
                        help="Complete path to the XML satellite schedule file.",
                        required=True)
    parser.add_argument("-o", "--output_dir",
                        dest="output_dir",
                        help="Complete path to the XML satellite schedule file.",
                        default="./")
    parser.add_argument("-v", "--verbose",
                        dest="verbosity",
                        action="count",
                        default=0,
                        help="Verbosity (between 1 and 2 occurrences with more leading to more "
                        "verbose logging). WARN=0, INFO=1, "
                        "DEBUG=2. This is overridden by the log config file if specified.")

    return parser.parse_args(args)


def run(args=None):
    """Create shapefiles from an XML schedule of satellite passes."""
    opts = parse_args(args)

    setup_logging(opts)
    filename = opts.xml_filepath
    satellites = opts.list_of_satellites
    tle_filename = opts.tle_filepath
    outdir = opts.output_dir

    shapefiles_from_schedule_xml_requests(filename,
                                          satellites,
                                          tle_filename,
                                          outdir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013, 2014, 2015, 2018 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Some generalized spherical functions.

base type is a numpy array of size (n, 2) (2 for lon and lats)

"""

import logging

import numpy as np
import pyresample.spherical

logger = logging.getLogger(__name__)

EPSILON = 0.0000001


def modpi(val, mod=np.pi):
    """Put *val* between -*mod* and *mod*."""
    return (val + mod) % (2 * mod) - mod


class SphPolygon(pyresample.spherical.SphPolygon):
    """A spherical polygon with drawing capabilities."""

    def draw(self, mapper, options, **more_options):
        """Draw the polygon."""
        lons = np.rad2deg(self.lon.take(np.arange(len(self.lon) + 1),
                                        mode="wrap"))
        lats = np.rad2deg(self.lat.take(np.arange(len(self.lat) + 1),
                                        mode="wrap"))
        rx, ry = mapper(lons, lats)
        mapper.plot(rx, ry, options, **more_options)


def get_twilight_poly(utctime):
    """Return a polygon enclosing the sunlit part of the globe at *utctime*."""
    from pyorbital import astronomy
    ra, dec = astronomy.sun_ra_dec(utctime)
    lon = modpi(ra - astronomy.gmst(utctime))
    lat = dec

    vertices = np.zeros((4, 2))

    vertices[0, :] = modpi(lon - np.pi / 2), 0
    if lat <= 0:
        vertices[1, :] = lon, np.pi / 2 + lat
        vertices[3, :] = modpi(lon + np.pi), -(np.pi / 2 + lat)
    else:
        vertices[1, :] = modpi(lon + np.pi), np.pi / 2 - lat
        vertices[3, :] = lon, -(np.pi / 2 - lat)

    vertices[2, :] = modpi(lon + np.pi / 2), 0

    return SphPolygon(vertices)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c22526.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The log handling."""

import logging
import logging.config
import logging.handlers

import yaml

LOG_FORMAT = "[%(asctime)s %(levelname)-8s] %(message)s"

log_levels = {
    0: logging.WARN,
    1: logging.INFO,
    2: logging.DEBUG,
}


def setup_logging(cmd_args):
    """Set up logging."""
    if cmd_args.log_config is not None:
        with open(cmd_args.log_config) as fd:
            log_dict = yaml.safe_load(fd.read())
            logging.config.dictConfig(log_dict)
            return

    root = logging.getLogger("")
    root.setLevel(log_levels[cmd_args.verbosity])

    fh_ = logging.StreamHandler()

    formatter = logging.Formatter(LOG_FORMAT)
    fh_.setFormatter(formatter)

    root.addHandler(fh_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024 Pytroll developers

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for satellite pass and schedule calculations."""


class SatScore:
    """Container for the score parameter deciding which satellite pass to take."""

    def __init__(self, day, night):
        """Initialize the score."""
        self.day = day
        self.night = night


class Satellite:
    """The Satellite class holding information on its score, name and designator."""

    def __init__(self, name, day, night,
                 schedule_name=None, international_designator=None):
        """Initialize the satellite."""
        self.name = name
        self.international_designator = international_designator
        self.score = SatScore(day, night)
        self.schedule_name = schedule_name or name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 - 2024 Pytroll

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the schedule module."""
import os
import tempfile
from datetime import datetime, timedelta
from unittest.mock import patch

import pytest
import yaml

from trollsched.satpass import get_aqua_terra_dumps, get_metopa_passes, get_next_passes
from trollsched.schedule import build_filename, conflicting_passes, fermia, fermib, get_passes_from_xml_file, run


class TestTools:
    """Test the tools."""

    def test_conflicting_passes(self):
        """Test conflicting passes."""
        class MyPass(object):

            def __init__(self, rise, fall):
                self.risetime = rise
                self.falltime = fall

        ref_time = datetime.utcnow()
        passes = [MyPass(ref_time, ref_time + timedelta(minutes=10)),
                  MyPass(ref_time + timedelta(minutes=10.01),
                         ref_time + timedelta(minutes=20))]
        assert len(conflicting_passes(passes, timedelta(seconds=0))) == 2
        assert len(conflicting_passes(passes, timedelta(seconds=60))) == 1


class TestUtils:
    """Test class for utilities."""

    def test_fermi(self):
        """Test the fermi formula."""
        assert fermia(0.25) == 0.875
        assert fermib(0.25) == 0.5

    def test_build_filename(self):
        """Test building filename."""
        tempdir = tempfile.gettempdir()
        pattern_name = "dir_output"
        pattern_dict = {"file_xml": os.path.join("{dir_output}",
                                                 "{date}-{time}-aquisition-schedule-{mode}-{station}.xml"),
                        "file_sci": os.path.join("{dir_output}", "scisys-schedule-{station}.txt"),
                        "dir_plots": os.path.join("{dir_output}", "plots.{station}"), "dir_output": tempdir,
                        "file_graph": os.path.join("{dir_output}", "graph.{station}")}
        kwargs = {"date": "20190104", "output_dir": ".", "dir_output": tempdir, "time": "122023"}

        res = build_filename(pattern_name, pattern_dict, kwargs)
        assert res == tempdir

        pattern_name = "file_xml"
        kwargs = {"station": "nrk", "mode": "request", "time": "125334",
                  "date": "20190104", "dir_output": tempdir, "output_dir": "."}
        res = build_filename(pattern_name, pattern_dict, kwargs)
        assert res == os.path.join(tempdir, "20190104-125334-aquisition-schedule-request-nrk.xml")


class TestAll:
    """The test class."""

    def setup_method(self):
        """Set up."""
        from pyorbital import orbital

        from trollsched.pass_scheduling_utils import Satellite

        self.utctime = datetime(2018, 11, 28, 10, 0)
        self.satellites = ["noaa-20", ]
        self.tles = {"noaa-20": {}}
        self.tles["noaa-20"]["line1"] = "1 43013U 17073A   18331.00000000  .00000048  00000-0  22749-4 0  3056"
        self.tles["noaa-20"]["line2"] = "2 43013 098.7413 267.0121 0001419 108.5818 058.1314 14.19552981053016"

        self.aquas = ["aqua", ]
        self.terras = ["terra", ]
        self.terra = Satellite("terra", 0, 0)
        self.metopa = Satellite("metop-a", 0, 0)

        self.tles["aqua"] = {}
        self.tles["aqua"]["line1"] = "1 27424U 02022A   18332.21220389  .00000093  00000-0  30754-4 0  9994"
        self.tles["aqua"]["line2"] = "2 27424  98.2121 270.9368 0001045 343.9225 155.8703 14.57111538881313"
        self.tles["terra"] = {}
        self.tles["terra"]["line1"] = "1 25994U 99068A   18338.20920286  .00000076  00000-0  26867-4 0  9999"
        self.tles["terra"]["line2"] = "2 25994  98.2142  50.5750 0000577 102.5211 257.6060 14.57132862  8586"
        self.tles["metop-a"] = {}
        self.tles["metop-a"]["line1"] = "1 29499U 06044A   18338.30873671  .00000000  00000+0  31223-4 0 00013"
        self.tles["metop-a"]["line2"] = "2 29499  98.6045  31.7725 0001942  91.8780 346.4884 14.21536046629175"

        self.orb = orbital.Orbital("NOAA 20",
                                   line1=self.tles["noaa-20"]["line1"],
                                   line2=self.tles["noaa-20"]["line2"])
        self.aqua_orb = orbital.Orbital("AQUA",
                                        line1=self.tles["aqua"]["line1"],
                                        line2=self.tles["aqua"]["line2"])
        self.terra_orb = orbital.Orbital("TERRA",
                                         line1=self.tles["terra"]["line1"],
                                         line2=self.tles["terra"]["line2"])
        self.metopa_orb = orbital.Orbital("Metop-A",
                                          line1=self.tles["metop-a"]["line1"],
                                          line2=self.tles["metop-a"]["line2"])

        # These values were used to generate the get_next_passes list mock:
        # utctime = datetime(2018, 12, 4, 9, 0)
        # forward = 6
        # coords = (16, 58, 0)
        self.metopa_passlist = [(datetime(2018, 12, 4, 9, 10, 4, 574801),
                                 datetime(2018, 12, 4, 9, 25, 29, 157194),
                                 datetime(2018, 12, 4, 9, 17, 48, 530484)),
                                (datetime(2018, 12, 4, 10, 50, 23, 899232),
                                 datetime(2018, 12, 4, 11, 4, 2, 335184),
                                 datetime(2018, 12, 4, 10, 57, 13, 691637)),
                                (datetime(2018, 12, 4, 12, 30, 24, 97160),
                                 datetime(2018, 12, 4, 12, 40, 42, 403698),
                                 datetime(2018, 12, 4, 12, 35, 33, 317647)),
                                (datetime(2018, 12, 4, 14, 9, 1, 937869),
                                 datetime(2018, 12, 4, 14, 17, 20, 556654),
                                 datetime(2018, 12, 4, 14, 13, 11, 247497))]

        self.dumpdata = [
            {"los": datetime(2018, 11, 28, 10, 0, 30), "station": "USAK05",
             "aos": datetime(2018, 11, 28, 9, 50, 24), "elev": "11.188"},
            {"los": datetime(2018, 11, 28, 11, 39, 47), "station": "AS2",
             "aos": datetime(2018, 11, 28, 11, 28, 51), "elev": "39.235"},
            {"los": datetime(2018, 11, 28, 13, 19, 8), "station": "USAK05",
             "aos": datetime(2018, 11, 28, 13, 6, 36), "elev": "58.249"},
            {"los": datetime(2018, 11, 28, 14, 54, 25), "station": "AS2",
             "aos": datetime(2018, 11, 28, 14, 44, 37), "elev": "22.403"},
            {"los": datetime(2018, 11, 28, 16, 27, 22), "station": "SG1",
             "aos": datetime(2018, 11, 28, 16, 16, 58), "elev": "9.521"}
        ]
        self.dumpdata_terra = [{"los": datetime(2018, 11, 20, 23, 24, 41), "station": "SG2",
                                "aos": datetime(2018, 11, 20, 23, 12, 32), "elev": "17.4526"},
                               {"los": datetime(2018, 11, 22, 23, 19, 21), "station": "AS3",
                                "aos": datetime(2018, 11, 22, 23, 8, 55), "elev": "28.9558"},
                               {"los": datetime(2018, 11, 22, 23, 19, 21), "station": "AS3",
                                "aos": datetime(2018, 11, 22, 23, 8, 55), "elev": "28.9558"},
                               {"los": datetime(2018, 11, 26, 22, 47, 34), "station": "SG1",
                                "aos": datetime(2018, 11, 26, 22, 34, 58), "elev": "21.5694"},
                               {"los": datetime(2018, 11, 26, 22, 47, 34), "station": "SG1",
                                "aos": datetime(2018, 11, 26, 22, 34, 58), "elev": "21.5694"},
                               {"los": datetime(2018, 11, 26, 22, 47, 34), "station": "SG1",
                                "aos": datetime(2018, 11, 26, 22, 34, 58), "elev": "21.5694"},
                               {"los": datetime(2018, 11, 27, 23, 30, 44), "station": "SG2",
                                "aos": datetime(2018, 11, 27, 23, 18, 39), "elev": "16.8795"},
                               {"los": datetime(2018, 11, 27, 23, 30, 44), "station": "SG2",
                                "aos": datetime(2018, 11, 27, 23, 18, 39), "elev": "16.8795"},
                               {"los": datetime(2018, 11, 28, 22, 43, 53), "station": "USAK05",
                                "aos": datetime(2018, 11, 28, 22, 31, 57), "elev": "40.9264"},
                               {"los": datetime(2018, 11, 28, 22, 43, 53), "station": "USAK05",
                                "aos": datetime(2018, 11, 28, 22, 31, 57), "elev": "40.9264"},
                               {"los": datetime(2018, 11, 29, 23, 25, 11), "station": "USAK05",
                                "aos": datetime(2018, 11, 29, 23, 14, 47), "elev": "26.9937"},
                               {"los": datetime(2018, 11, 29, 23, 25, 11), "station": "USAK05",
                                "aos": datetime(2018, 11, 29, 23, 14, 47), "elev": "26.9937"},
                               {"los": datetime(2018, 11, 30, 22, 31, 3), "station": "AS2",
                                "aos": datetime(2018, 11, 30, 22, 19, 48), "elev": "47.8599"},
                               {"los": datetime(2018, 12, 1, 1, 29, 2), "station": "WG1",
                                "aos": datetime(2018, 12, 1, 1, 21, 11), "elev": "8.0543"},
                               {"los": datetime(2018, 11, 30, 22, 31, 3), "station": "AS2",
                                "aos": datetime(2018, 11, 30, 22, 19, 48), "elev": "47.8599"},
                               {"los": datetime(2018, 12, 1, 1, 29, 2), "station": "WG1",
                                "aos": datetime(2018, 12, 1, 1, 21, 11), "elev": "8.0543"},
                               {"los": datetime(2018, 12, 3, 1, 28, 14), "station": "SG2",
                                "aos": datetime(2018, 12, 3, 1, 17, 53), "elev": "9.2428"},
                               {"los": datetime(2018, 12, 3, 22, 53, 35), "station": "SG1",
                                "aos": datetime(2018, 12, 3, 22, 41, 5), "elev": "20.8371"},
                               {"los": datetime(2018, 12, 3, 22, 53, 35), "station": "SG1",
                                "aos": datetime(2018, 12, 3, 22, 41, 5), "elev": "20.8371"},
                               {"los": datetime(2018, 12, 4, 23, 43, 5), "station": "AS2",
                                "aos": datetime(2018, 12, 4, 23, 33, 8), "elev": "23.546"}]

    @patch("os.path.exists")
    def test_get_next_passes_viirs(self, exists):
        """Test getting the next viirs passes."""
        exists.return_code = True

        # mymock:
        with patch("pyorbital.orbital.Orbital") as mymock:
            instance = mymock.return_value
            instance.get_next_passes = self.orb.get_next_passes

            allpasses = get_next_passes(self.satellites, self.utctime,
                                        4, (16, 58, 0), tle_file="nonexisting")

            assert len(allpasses) == 2

            rt1 = datetime(2018, 11, 28, 10, 53, 42, 79483)
            ft1 = datetime(2018, 11, 28, 11, 9, 6, 916787)
            rt2 = datetime(2018, 11, 28, 12, 34, 44, 667963)
            ft2 = datetime(2018, 11, 28, 12, 49, 25, 134067)

            rise_times = [p.risetime for p in allpasses]
            fall_times = [p.falltime for p in allpasses]

            assert rt1 in rise_times
            assert rt2 in rise_times
            assert ft1 in fall_times
            assert ft2 in fall_times

            assert all([p.instrument == "viirs" for p in allpasses])

    @patch("os.path.exists")
    @patch("trollsched.satpass.get_aqua_terra_dumpdata_from_ftp")
    def test_get_next_passes_with_aquadumps(self, dumps_from_ftp, exists):
        """Test getting the passes with dumps."""
        dumps_from_ftp.return_value = self.dumpdata
        exists.return_code = True

        # mymock:
        with patch("pyorbital.orbital.Orbital") as mymock:
            instance = mymock.return_value
            instance.get_next_passes = self.aqua_orb.get_next_passes

            allpasses = get_next_passes(self.aquas, self.utctime,
                                        6, (16, 58, 0), tle_file="nonexisting",
                                        aqua_terra_dumps=True)

            assert len(allpasses) == 3

            rt1 = datetime(2018, 11, 28, 11, 12, 8, 728455)
            ft1 = datetime(2018, 11, 28, 11, 26, 8, 250021)
            rt2 = datetime(2018, 11, 28, 12, 50, 46, 574975)
            ft2 = datetime(2018, 11, 28, 13, 3, 53, 262440)
            rt3 = datetime(2018, 11, 28, 14, 33, 33, 973194)
            ft3 = datetime(2018, 11, 28, 14, 40, 10, 761405)

            for mypass in allpasses:
                dtmin = timedelta(seconds=10000000)
                for risetime in [rt1, rt2, rt3]:
                    dt_ = abs(mypass.risetime - risetime)
                    if dt_ < dtmin:
                        dtmin = dt_

                assert dtmin.seconds == pytest.approx(0)

                dtmin = timedelta(seconds=10000000)
                for falltime in [ft1, ft2, ft3]:
                    dt_ = abs(mypass.falltime - falltime)
                    if dt_ < dtmin:
                        dtmin = dt_

                assert dtmin.seconds == pytest.approx(0)

                assert mypass.instrument == "modis"

    @patch("trollsched.satpass.get_aqua_terra_dumpdata_from_ftp")
    def test_get_aqua_terra_dumps(self, dumps_from_ftp):
        """Test getting the EOS dumps."""
        dumps_from_ftp.return_value = self.dumpdata_terra

        # mymock:
        with patch("pyorbital.orbital.Orbital") as mymock:
            instance = mymock.return_value
            instance.get_next_passes = self.terra_orb.get_next_passes

            dumps = get_aqua_terra_dumps(datetime(2018, 12, 3, 0, 0),
                                         datetime(2018, 12, 10, 0, 0),
                                         self.terra_orb,
                                         self.terra)

            assert len(dumps) == 4
            assert dumps[0].station == "SG2"
            assert dumps[0].max_elev == "9.2428"
            assert dumps[0].pass_direction() == "ascending"
            assert (dumps[0].risetime - datetime(2018, 12, 3, 1, 17, 53)).seconds == 0
            assert (dumps[0].falltime - datetime(2018, 12, 3, 1, 28, 14)).seconds == 0

            assert dumps[3].station == "AS2"
            assert dumps[3].max_elev == "23.546"
            assert dumps[3].pass_direction() == "descending"
            assert (dumps[3].risetime - datetime(2018, 12, 4, 23, 33, 8)).seconds == 0
            assert (dumps[3].falltime - datetime(2018, 12, 4, 23, 43, 5)).seconds == 0

    @patch("os.path.exists")
    def test_get_metopa_passes(self, exists):
        """Test getting metopa passes."""
        exists.return_code = True

        with patch("pyorbital.orbital.Orbital") as mymock:
            instance = mymock.return_value
            instance.get_next_passes = self.metopa_orb.get_next_passes

            metopa_passes = get_metopa_passes(self.metopa, self.metopa_passlist, self.metopa_orb)

            assert len(metopa_passes) == 2
            assert metopa_passes[0].pass_direction() == "descending"
            assert metopa_passes[0].seconds() == pytest.approx(487.512589, 1e-5)
            assert (metopa_passes[0].uptime - datetime(2018, 12, 4, 9, 17, 48, 530484)).seconds == 0
            assert (metopa_passes[0].risetime - datetime(2018, 12, 4, 9, 17, 21, 644605)).seconds == 0


euron1 = """euron1:
  description: Northern Europe - 1km
  projection:
    proj: stere
    ellps: WGS84
    lat_0: 90.0
    lon_0: 0.0
    lat_ts: 60.0
  shape:
    height: 3072
    width: 3072
  area_extent:
    lower_left_xy: [-1000000.0, -4500000.0]
    upper_right_xy: [2072000.0, -1428000.0]
"""


def test_pyorbitals_platform_name(tmp_path):
    """Test that using pyorbital's platform name allows spurious names in the TLE data."""
    spurious_tle = ("NOAA 20 (JPSS-1)\n"
                    "1 43013U 17073A   24093.57357837  .00000145  00000+0  86604-4 0  9999\n"
                    "2 43013  98.7039  32.7741 0007542 324.8026  35.2652 14.21254587330172\n")

    config_file = tmp_path / "config.yaml"
    tle_file = tmp_path / "test.tle"
    area_file = tmp_path / "areas.yaml"
    sched_file = tmp_path / "mysched.xml"

    with open(area_file, "w") as fd:
        fd.write(euron1)

    with open(tle_file, "w") as fd:
        fd.write(spurious_tle)

    config = dict(default=dict(station=["nrk"],
                               forward=12,
                               start=0,
                               center_id="SMHI"),
                  stations=dict(nrk=dict(name="nrk",
                                         longitude=16,
                                         latitude=58,
                                         altitude=0,
                                         satellites=["noaa-20"],
                                         area="euron1",
                                         area_file=os.fspath(area_file))),

                  pattern=dict(dir_output=os.fspath(tmp_path),
                               file_xml=os.fspath(sched_file)),
                  satellites={"noaa-20": dict(schedule_name="noaa20",
                                              international_designator="43013",
                                              night=0.4,
                                              day=0.9)}
                  )

    with open(config_file, "w") as fd:
        fd.write(yaml.dump(config))

    run(["-c", os.fspath(config_file), "-x", "-t", os.fspath(tle_file)])
    assert sched_file in tmp_path.iterdir()


avoid = """<?xml version="1.0"?>
<acquisition-schedule>
  <pass satellite="suomi npp" start-time="2024-05-08-03:57:01" end-time="2024-05-08-04:09:35"/>
</acquisition-schedule>
"""


def test_schedule_avoid(tmp_path):
    """Test that schedule can handle avoid list."""
    tle = ("SUOMI NPP\n"
           "1 37849U 11061A   24128.72979065  .00000000  00000+0  12832-3 0 00014\n"
           "2 37849  98.7205  67.3215 0001498  76.6175 303.3589 14.19560637649138\n"
           "NOAA 20\n"
           "1 37849U 11061A   24128.72979065  .00000000  00000+0  12832-3 0 00014\n"
           "2 37849  98.7205  67.3215 0001498  76.6175 303.3589 14.19560637649138\n")

    config_file = tmp_path / "config.yaml"
    tle_file = tmp_path / "test.tle"
    area_file = tmp_path / "areas.yaml"
    sched_file = tmp_path / "sched-with-avoid.xml"
    avoid_file = tmp_path / "avoid.xml"

    with open(area_file, "w") as fd:
        fd.write(euron1)

    with open(tle_file, "w") as fd:
        fd.write(tle)

    with open(avoid_file, "w") as fd:
        fd.write(avoid)

    config = dict(default=dict(station=["nrk"],
                               forward=12,
                               start=0,
                               center_id="SMHI"),
                  stations=dict(nrk=dict(name="nrk",
                                         longitude=16,
                                         latitude=58,
                                         altitude=0,
                                         satellites=["suomi npp", "noaa 20"],
                                         area="euron1",
                                         area_file=os.fspath(area_file))),

                  pattern=dict(dir_output=os.fspath(tmp_path),
                               file_xml=os.fspath(sched_file)),
                  satellites={"suomi npp": dict(schedule_name="suomi npp",
                                                international_designator="37849",
                                                night=0.4,
                                                day=0.9),
                              "noaa 20": dict(schedule_name="noaa 20",
                                              international_designator="99999",
                                              night=0.4,
                                              day=0.9)}
                  )

    with open(config_file, "w") as fd:
        fd.write(yaml.dump(config))

    start_time = datetime(2024, 5, 8, 0, 0, 0)
    run(["-c", os.fspath(config_file), "-x", "-v", "-t", os.fspath(tle_file),
         "--start-time", start_time.strftime("%Y-%m-%dT%H:%M:%S"), "--avoid", os.fspath(avoid_file)])
    assert sched_file in tmp_path.iterdir()

    sched_file_passes = get_passes_from_xml_file([sched_file])
    avoid_file_passes = get_passes_from_xml_file([avoid_file])
    for avoid_pass in avoid_file_passes:
        assert avoid_pass not in sched_file_passes



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016, 2018, 2019 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""From a schedule request xml file generate png's with swath coverage outline and an xml page for visualisation.

It uses posttroll to listen for incoming schedule request xml files and then triggers the png and xml output generation.
"""

import logging
import os.path
import sys
from datetime import datetime

import defusedxml.ElementTree as ET
import posttroll.subscriber
from posttroll.publisher import Publish
from six.moves.configparser import RawConfigParser
from six.moves.urllib.parse import urlparse

from trollsched import INSTRUMENT, SATELLITE_NAMES
from trollsched.drawing import save_fig
from trollsched.satpass import Pass

LOG = logging.getLogger(__name__)

CFG_DIR = os.environ.get("PYTROLL_SCHEDULE_CONFIG_DIR", "./")
CONF = RawConfigParser()
CFG_FILE = os.path.join(CFG_DIR, "pytroll_schedule_config.cfg")
LOG.debug("Config file = " + str(CFG_FILE))
if not os.path.exists(CFG_FILE):
    raise IOError("Config file %s does not exist!" % CFG_FILE)

CONF.read(CFG_FILE)
OPTIONS = {}
for option, value in CONF.items("DEFAULT"):
    OPTIONS[option] = value


#: Default time format
_DEFAULT_TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

#: Default log format
_DEFAULT_LOG_FORMAT = "[%(levelname)s: %(asctime)s : %(name)s] %(message)s"


def process_xmlrequest(filename, plotdir, output_file, excluded_satellites):
    """Process the xml request."""
    tree = ET.parse(filename)
    root = tree.getroot()

    for child in root:
        if child.tag == "pass":
            LOG.debug("Pass: %s", str(child.attrib))
            platform_name = SATELLITE_NAMES.get(child.attrib["satellite"], child.attrib["satellite"])
            instrument = INSTRUMENT.get(platform_name)
            if not instrument:
                LOG.error("Instrument unknown! Platform = %s", platform_name)
                continue

            if platform_name in excluded_satellites:
                LOG.debug("Platform name excluded: %s", platform_name)
                continue
            try:
                overpass = Pass(platform_name,
                                datetime.strptime(child.attrib["start-time"],
                                                  "%Y-%m-%d-%H:%M:%S"),
                                datetime.strptime(child.attrib["end-time"],
                                                  "%Y-%m-%d-%H:%M:%S"),
                                instrument=instrument)
            except KeyError as err:
                LOG.warning("Failed on satellite %s: %s", platform_name, str(err))
                continue

            save_fig(overpass, directory=plotdir)
            child.set("img", overpass.fig)
            child.set("rec", "True")
            LOG.debug("Plot saved - plotdir = %s, platform_name = %s", plotdir, platform_name)

    tree.write(output_file, encoding="utf-8", xml_declaration=True)

    with open(output_file) as fpt:
        lines = fpt.readlines()
        lines.insert(
            1, "<?xml-stylesheet type='text/xsl' href='reqreader.xsl'?>")

    with open(output_file, "w") as fpt:
        fpt.writelines(lines)


def start_plotting(jobreg, message, **kwargs):
    """Make a web-publishable version of the xml schedule.

    Read the xmlschedule request file and make the png images of swath outlines
    and generate the output xml file for web publication.
    """
    excluded_satellites = kwargs.get("excluded_satellites", [])

    LOG.info("")
    LOG.info("job-registry dict: " + str(jobreg))
    LOG.info("\tMessage:")
    LOG.info(message)
    urlobj = urlparse(message.data["uri"])
    # path, fname = os.path.split(urlobj.path)

    process_xmlrequest(urlobj.path,
                       OPTIONS["path_plots"], OPTIONS["xmlfilepath"],
                       excluded_satellites)

    return jobreg


def schedule_page_generator(excluded_satellite_list=None):
    """Listens and triggers processing."""
    LOG.info(
        "*** Start the generation of the schedule xml page with swath outline plots")
    with posttroll.subscriber.Subscribe("", [OPTIONS["posttroll_topic"], ],
                                        True) as subscr:
        with Publish("schedule_page_generator", 0) as publisher:
            job_registry = {}
            for msg in subscr.recv():
                job_registry = start_plotting(
                    job_registry, msg, publisher=publisher, excluded_satellites=excluded_satellite_list)
                # Cleanup in registry (keep only the last 5):
                keys = job_registry.keys()
                if len(keys) > 5:
                    keys.sort()
                    job_registry.pop(keys[0])


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--excluded_satellites", nargs="*",
                        help="List of platform names to exclude",
                        default=[])
    opts = parser.parse_args()

    no_sats = opts.excluded_satellites

    handler = logging.StreamHandler(sys.stderr)

    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(fmt=_DEFAULT_LOG_FORMAT,
                                  datefmt=_DEFAULT_TIME_FORMAT)
    handler.setFormatter(formatter)
    logging.getLogger("").addHandler(handler)
    logging.getLogger("").setLevel(logging.DEBUG)
    logging.getLogger("posttroll").setLevel(logging.INFO)

    LOG = logging.getLogger("schedule_page_generator")
    LOG.info("Exclude the following satellite platforms: %s", str(no_sats))

    schedule_page_generator(no_sats)

    # uri = "/data/temp/AdamD/xxx/2018-10-22-00-42-28-acquisition-schedule-confirmation-nrk.xml"
    # urlobj = urlparse(uri)
    # process_xmlrequest(urlobj.path,
    #                    OPTIONS['path_plots'], OPTIONS['xmlfilepath'], no_sats)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018, 2024 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for the pytroll-schedule methods.

E.g. nightshade method for Cartopy as available in Basemap.
"""

from datetime import datetime

import numpy as np


def sun_pos(dt=None):
    """Get location on earth where the sun is in zenith at time *dt*.

    This function computes a rough estimate of the coordinates for
    the point on the surface of the Earth where the Sun is directly
    overhead at the time dt. Precision is down to a few degrees. This
    means that the equinoxes (when the sign of the latitude changes)
    will be off by a few days.

    The function is intended only for visualization. For more precise
    calculations consider for example the PyEphem package.

    Taken from here:
    https://scitools.org.uk/cartopy/docs/latest/gallery/aurora_forecast.html

    Parameters
    ----------
    dt: datetime
        Defaults to datetime.utcnow()

    Returns:
    -------
    lat, lng: tuple of floats
        Approximate coordinates of the point where the sun is
        in zenith at the time dt.

    """
    if dt is None:
        dt = datetime.utcnow()

    axial_tilt = 23.4
    ref_solstice = datetime(2016, 6, 21, 22, 22)
    days_per_year = 365.2425
    seconds_per_day = 24 * 60 * 60.0

    days_since_ref = (dt - ref_solstice).total_seconds() / seconds_per_day
    lat = axial_tilt * np.cos(2 * np.pi * days_since_ref / days_per_year)
    sec_since_midnight = (dt - datetime(dt.year, dt.month, dt.day)).seconds
    lng = -(sec_since_midnight / seconds_per_day - 0.5) * 360
    return lat, lng


def fill_dark_side(ax, time=None, **kwargs):
    """Plot a fill on the dark side of the planet (without refraction).

    Parameters
    ----------
        ax : Matplotlib axes
            The axes to plot on.
        time : datetime
            The time to calculate terminator for. Defaults to datetime.utcnow()
        **kwargs :
            Passed on to Matplotlib's ax.fill()

    """
    import cartopy.crs as ccrs

    lat, lng = sun_pos(time)
    pole_lng = lng
    if lat > 0:
        pole_lat = -90 + lat
        central_rot_lng = 180
    else:
        pole_lat = 90 + lat
        central_rot_lng = 0

    rotated_pole = ccrs.RotatedPole(pole_latitude=pole_lat,
                                    pole_longitude=pole_lng,
                                    central_rotated_longitude=central_rot_lng)

    x = np.empty(360)
    y = np.empty(360)
    x[:180] = -90
    y[:180] = np.arange(-90, 90.)
    x[180:] = 90
    y[180:] = np.arange(90, -90., -1)

    ax.fill(x, y, transform=rotated_pole, **kwargs)



"""Writers for different schedule formats."""
import os
from datetime import datetime


def generate_meos_file(output_file, allpasses, coords, start, report_mode=False):
    """Generate a meos file."""
    with open(output_file, "w") as out:
        out.write(" No. Date    Satellite  Orbit Max EL  AOS      Ovlp  LOS      Durtn  Az(AOS/MAX)\n")
        line_no = 1
        for overpass in sorted(allpasses, key=lambda x: x.risetime):
            if (overpass.rec or report_mode) and overpass.risetime > start:
                out.write(overpass.print_meos(coords, line_no) + "\n")
                line_no += 1
        out.close()
    return output_file


def generate_sch_file(output_file, overpasses, coords):
    """Generate a vcs/scisys/cgi schedule file."""
    with open(output_file, "w") as out:
        # create epochs
        out.write("#Orbital elements\n#\n#SCName           Epochtime\n#\n")
        satellites = set()

        for overpass in overpasses:
            epoch = "!{0:<16} {1}".format(overpass.satellite.name.upper(),
                                          overpass.orb.tle.epoch.strftime("%Y%m%d %H%M%S"))
            satellites |= set([epoch])
        sats = "\n".join(satellites) + "\n"
        out.write(sats)
        out.write("#\n#\n#Pass List\n#\n")

        out.write(
            "#SCName          RevNum Risetime        Falltime        Elev Dura ANL   Rec Dir Man Ovl OvlSCName        "
            "OvlRev OvlRisetime     OrigRisetime    OrigFalltime    OrigDuration\n#\n")

        for overpass in sorted(overpasses):
            out.write(overpass.print_vcs(coords) + "\n")


def generate_metno_xml_file(output_file, allpasses, coords, start, end, station_name, center_id, report_mode=False):
    """Generate a meto xml file."""
    import defusedxml.ElementTree as ET

    reqtime = datetime.utcnow()
    time_format = "%Y-%m-%dT%H:%M:%S"

    with open(output_file, "w") as out:
        out.write("<?xml version='1.0' encoding='utf-8'?>")

        root = ET.Element("acquisition-schedule")
        props = ET.SubElement(root, "properties")
        proj = ET.SubElement(props, "project")
        proj.text = "Pytroll"
        typep = ET.SubElement(props, "type")
        if report_mode:
            typep.text = "report"
        else:
            typep.text = "request"
        station = ET.SubElement(props, "station")
        station.text = station_name
        file_start = ET.SubElement(props, "file-start")
        file_start.text = start.strftime(time_format)
        file_end = ET.SubElement(props, "file-end")
        file_end.text = end.strftime(time_format)
        reqby = ET.SubElement(props, "requested-by")
        reqby.text = center_id
        reqon = ET.SubElement(props, "requested-on")
        reqon.text = reqtime.strftime(time_format)

        for overpass in sorted(allpasses, key=lambda x: x.risetime):
            if (overpass.rec or report_mode) and overpass.risetime > start:
                overpass.generate_metno_xml(coords, root)

        out.write(ET.tostring(root).decode("utf-8"))
        out.close()
    return output_file


def generate_xml_requests(sched, start, end, station_name, center_id, report_mode=False):
    """Create xml requests."""
    # defusedxml is not needed here as we only generate an xml file (ie no reading of potentially harmful data)
    from xml.etree import ElementTree as ET  # noqa

    reqtime = datetime.utcnow()
    time_format = "%Y-%m-%d-%H:%M:%S"

    root = ET.Element("acquisition-schedule")
    props = ET.SubElement(root, "properties")
    proj = ET.SubElement(props, "project")
    proj.text = "Pytroll"
    typep = ET.SubElement(props, "type")
    if report_mode:
        typep.text = "report"
    else:
        typep.text = "request"
    station = ET.SubElement(props, "station")
    station.text = station_name
    file_start = ET.SubElement(props, "file-start")
    file_start.text = start.strftime(time_format)
    file_end = ET.SubElement(props, "file-end")
    file_end.text = end.strftime(time_format)
    reqby = ET.SubElement(props, "requested-by")
    reqby.text = center_id
    reqon = ET.SubElement(props, "requested-on")
    reqon.text = reqtime.strftime(time_format)

    for overpass in sorted(sched):
        if (overpass.rec or report_mode) and overpass.risetime > start:
            ovpass = ET.SubElement(root, "pass")
            sat_name = overpass.satellite.schedule_name or overpass.satellite.name
            ovpass.set("satellite", sat_name)
            ovpass.set("start-time", overpass.risetime.strftime(time_format))
            ovpass.set("end-time", overpass.falltime.strftime(time_format))
            if report_mode:
                if overpass.fig is not None:
                    ovpass.set("img", overpass.fig)
                ovpass.set("rec", str(overpass.rec))

    return root, reqtime


def generate_xml_file(sched, start, end, xml_file, station, center_id, report_mode=False):
    """Create an xml request file."""
    import defusedxml.ElementTree as ET
    tree, reqtime = generate_xml_requests(sched,
                                          start, end,
                                          station, center_id, report_mode)
    filename = xml_file
    tmp_filename = xml_file + reqtime.strftime("%Y-%m-%d-%H-%M-%S") + ".tmp"
    with open(tmp_filename, "w") as fp_:
        if report_mode:
            fp_.write("<?xml version='1.0' encoding='utf-8'?>"
                      "<?xml-stylesheet type='text/xsl' href='reqreader.xsl'?>")
        fp_.write(ET.tostring(tree).decode("utf-8"))
    os.rename(tmp_filename, filename)
    return filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, 2018, 2024 Martin Raspaud
#
# Author(s):
#
#   Alexander Maul <alexander.maul@dwd.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Combine several graphs."""
import logging
import sys
from datetime import datetime, timedelta

from trollsched.graph import Graph

logger = logging.getLogger("trollsched")


def add_graphs(graphs, passes, delay=timedelta(seconds=0)):
    """Add all graphs to one combined graph."""
    statlst = graphs.keys()

    def count_neq_passes(pl):
        """Counts how many satellite passes in a list are really distinct (satellite/epoch)."""
        if len(pl):
            r = []
            s = 1
            for q in pl[1:]:
                if pl[0] != q:
                    r.append(q)
            s += count_neq_passes(r)
            return s
        else:
            return 0

    for s, g in graphs.items():
        logger.debug("station: %s, order: %d", s, g.order)

    # Graphs and allpasses are hashmaps of sets, or similar, but we need
    # lists of lists, forthat they are copied.
    grl = []
    pl = []
    for s in statlst:
        grl.append(graphs[s])
        pl.append(sorted(passes[s], key=lambda x: x.risetime))

    # Rough estimate for the size of the combined passes' graph.
    n_vertices = 1
    for g in grl:
        n_vertices += g.order
    n_vertices *= len(statlst) * 2
    newgraph = Graph(n_vertices=n_vertices)

    logger.debug("newgraph order: %d", newgraph.order)

    # This value signals the end, when no more passes from any antenna are available.
    stopper = tuple((None, None) for s in range(len(statlst)))

    # The new passes list, it'll be filled with tuples, each with of one pass per antenna.
    # It's initialized with the first passes.
    #
    # TODO: ideally something like next line, but this doesn't work faultless
    # if one or more stations have multiple "first passes":
    # newpasses = [tuple((pl[s][p - 1], None) for s in range(len(statlst)) for p in grl[s].neighbours(0))]
    #
    # TODO: not "just the first vertix" with the line:
    # parlist = [newpasses[0]]
    #
    newpasses = [tuple((pl[s][grl[s].neighbours(0)[0] - 1], None) for s in range(len(statlst)))]
    parlist = [newpasses[0]]
    while len(parlist):
        newparlist = []
        for parnode in parlist:
            if parnode == stopper:
                # All antennas reached the end of passes list in this path of
                # possibilities.
                # stopper == ((None,None) * stations)
                #
                # If this happens for all elements of parlist, newparlist will
                # stay empty and (at the bottom of this loop) replace parlist,
                # which as an empty list will cause the surrounding while-loop
                # to end.
                continue

            collected_newnodes = collect_nodes(0, parnode, grl, newgraph, newpasses, pl, delay)

            for newnode_list in collected_newnodes:
                newnode = tuple(newnode_list)
                if newnode not in newpasses:
                    newpasses.append(newnode)

                if newnode not in newparlist:
                    newparlist.append(newnode)

                # Collecting the weights from each stations weight-matrix ...
                # (could be more compact if it weren't for the None-values)
                wl = []
                for s, p, n in zip(range(len(statlst)), parnode, newnode):
                    try:
                        if n[0] is None:
                            wl.append(0)
                        else:
                            wl.append(n[1] or grl[s].weight(pl[s].index(p[0]) + 1, pl[s].index(n[0]) + 1))
                    except Exception:
                        logger.error(
                            "Collecting weights: stat %d - parnode %s %s - newnode %s %s",
                            s, parnode, p, newnode, n, exc_info=1)
                        raise
                # Apply vertix-count to the sum of collected weights.
                # vertix-count: number of vertices with reference to same
                # satellite pass, it can result to 0, 1, 2.
                w = sum(wl) / 2 ** ((2 * len(parnode)) - count_neq_passes(parnode) - count_neq_passes(newnode))

                # TODO: if the starting point isn't "just the first vertix",
                # the comparison must be changed
                if parnode == newpasses[0]:
                    # "virtual" weight for the starting point.
                    newgraph.add_arc(0, newpasses.index(parnode) + 1, w)

                newgraph.add_arc(newpasses.index(parnode) + 1, newpasses.index(newnode) + 1, w)

        parlist = newparlist

    logger.debug("newpasses length: %d", len(newpasses))

    return statlst, newgraph, newpasses


def collect_nodes(statnr, parnode, graph_set, newgraph, newpasses, passes_list, delay=timedelta(seconds=0)):
    """Collect all nodes reachable from the nodes in parnode, creating all combinations.

    RETURN: [[a1, b1], [a1, b2], ..., [a2, b1], ...]
    """
    # All collected nodes are virtually occuring at the same time, so some nodes
    # might be pulled up in the timeline to create a set "overlapping" passes.
    # If there are no more passes available for one station, None is set.

    bufflist = []
    p = parnode[statnr]
    g = graph_set[statnr]

    def overlap_any(this, test_list):
        """Tests if this overlapps any of the new-nodes in test_list.

        The new-nodes are in form (vertix, simulated-weight), only nodes without
        simulated weight are considered in the test.

        RETURN: -1 | 0 | +1 , if this lies before, overlapps any, or lies after
        the nodes in test_list.
        """
        if len(test_list) == 0:
            return 0
        minrise = datetime.utcnow() + timedelta(days=300)
        maxfall = datetime.utcnow() - timedelta(days=300)
        for p in test_list:
            if p[0] is None or p[1] is not None:
                continue
            if p[0].risetime < minrise:
                minrise = p[0].risetime
            if p[0].falltime > maxfall:
                maxfall = p[0].falltime
        if minrise > maxfall:
            return 0
        elif this.falltime < minrise:
            return -1
        elif this.risetime > maxfall:
            return +1
        else:
            return 0

    if p == (None, None):
        # There won't be any collectable nodes.
        # This None will act as a filler in the combined-vertices-tuples,
        # to get the access-by-index right.
        gn = [None]

    elif p[1] is not None:
        # A simulated parent node is set as neighbours' list.
        # It'll be processed as if it's the node which occurs in this
        # time-slot -- which it propably does, otherwise it's subjected
        # to simulation (again!).
        gn = [passes_list[statnr].index(p[0]) + 1]

    else:
        # Special cases aside, this creates a list of neighbours to the
        # current passes node.
        try:
            gn = g.neighbours(passes_list[statnr].index(p[0]) + 1)
        except Exception:
            logger.debug(f"len(passes_list)={len(passes_list)}  len(graph_set)={len(graph_set)} statnr={statnr} p={p}")
            logger.debug("passes_list", passes_list)
            raise

        if gn[0] > len(passes_list[statnr]):
            # But if there weren't any neighbours, set an empty list.
            gn = [None]

    if statnr + 1 == len(parnode):
        # It's the 'rightmost' of the list parnode,
        # and the deepest point of the recursion.

        if None in gn:
            # That's "no further connection".
            # It get's a special treatment, because there is no None in the
            # passes-list we could access by index.
            bufflist = [[(None, None)]]

        else:
            # Prepare to return just the list of neighbouring vertices.
            for m in zip((passes_list[statnr][n - 1] for n in gn), (None for _ in gn)):
                bufflist.append([m])

    else:
        # Since it's not the last element of the list parnode, we recurse and
        # then permutade all vertix-lists together.
        col = collect_nodes(statnr + 1, parnode, graph_set, newgraph, newpasses, passes_list)

        # Creating the permutation of all neighbours with the list returned
        # by the recursion.
        # A simulated parent node is seen as a regular list of neighbours.
        for n in gn:
            for cx in col:
                try:
                    if n is None:
                        # The end-of-neighbours dummy.
                        cc = cx[:]
                        cc.insert(0, (None, None))
                        bufflist.append(cc)

                    else:
                        # Are two passes are overlapping?
                        overlap = overlap_any(passes_list[statnr][n - 1], cx)

                        if overlap == 0:
                            # Two passes overlapping, no special handling required.
                            cc = cx[:]
                            cc.insert(0, (
                                (passes_list[statnr][n - 1], None)
                            ))
                            bufflist.append(cc)

                        elif overlap > 0:
                            # If the current parent node's pass is not overlapping
                            # but AFTER the pass from the recursion-return-list
                            # the current parent node gets "simulated".
                            cc = cx[:]
                            cc.insert(0, (
                                passes_list[statnr][n - 1],
                                g.weight(passes_list[statnr].index(p[0]) + 1, n)
                            ))
                            bufflist.append(cc)

                        else:
                            # If the current parent node's pass is not overlapping
                            # but BEFORE the pass from the recursion-return-list
                            # the recursion-list-node gets "simulated".
                            cc = [
                                (c[0], graph_set[s].weight(
                                 passes_list[s].index(parnode[s][0]) + 1,
                                 passes_list[s].index(c[0]) + 1
                                 )
                                 ) if c != (None, None) else (None, None)
                                for s, c in zip(range(statnr + 1, len(parnode)), cx)
                            ]
                            cc.insert(0, (passes_list[statnr][n - 1], None))
                            bufflist.append(cc)
                except Exception:
                    logger.debug(f"\nCATCH\ngn:{gn}->{n}, col: {col}->{cx} statnr={statnr} statnr+i: {statnr + 1}")
                    logger.debug(f"len(passes_list -n -cx): {len(passes_list[statnr])} {len(passes_list[statnr + 1])}")
                    for s in range(statnr, len(passes_list)):
                        logger.debug(f"passes_list[{s}] => {passes_list[s]}")
                    raise

    return bufflist


def get_combined_sched(allgraphs, allpasses, delay_sec=60):

    delay = timedelta(seconds=delay_sec)

    statlst, newgraph, newpasses = add_graphs(allgraphs, allpasses, delay)

    # >>> DEV: test if the graphs could be "folded" to use use less RAM.
    # for s, g in allgraphs.items():
    #    print "test folding", s
    #    test_folding(g)
    # print "test folding newgraph"
    # if test_folding(newgraph):
    #    print_matrix(newgraph.adj_matrix, 25, 36)
    # <<<

    dist, path = newgraph.dag_longest_path(0, len(newpasses))

    logger.debug("Distance: %d", dist)
    logger.debug("Path through newpasses: %s", path)

    del dist
    return statlst, [newpasses[idx - 1] for idx in path[1:-1]], (newgraph, newpasses)


def print_matrix(m, ly=-1, lx=-1):
    """For DEBUG: Prints one of the graphs' backing matrix without flooding the screen.

    It'll print the first lx columns from the first ly rows, then
    the last lx columns from the last ly rows.
    """
    for i, l in zip(range(ly), m[0:ly]):
        logger.debug(f"{i}: {l[:lx]}...")
    logger.debug("[..., ...]")
    for i, l in zip(range(len(m) - ly - 1, len(m) - 1), m[-ly:]):
        logger.debug(f"{i}: ...{l[-lx:]}")

def test_folding(g):
    """Test if the graphs could be "folded".

    Test if the graphs could be "folded", or better "squished", to reduce
    size on cost of calculating the real x',y' from the x,y of the "unfolded"
    graph.
    """
    r = False
    for u in range(g.order):
        for n in g.neighbours(u):
            if n < u:
                logger.debug(f"{n}<{u}")
                r = True
    return r


def main():
    """The main script."""
    import logging
    import logging.handlers
    import os
    import pickle

    from trollsched.schedule import build_filename, combined_stations, parse_datetime

    try:
        import argparse

        from trollsched.schedule import read_config
        logger = logging.getLogger("trollsched")
        parser = argparse.ArgumentParser()
        parser.add_argument("-c", "--config", default=None,
                            help="configuration file to use")
        parser.add_argument("-s", "--start-time", type=parse_datetime,
                            help="start time of the schedule to compute")
        parser.add_argument("-o", "--output-dir", default=None,
                            help="where to put generated files")
        parser.add_argument("-x", "--xml", action="store_true",
                            help="generate an xml request file (schedule)"
                            )
        parser.add_argument("-r", "--report", action="store_true",
                            help="generate an xml report file (schedule)")
        parser.add_argument("--scisys", action="store_true",
                            help="generate a SCISYS schedule file")
        parser.add_argument("-p", "--plot", action="store_true",
                            help="generate plot images")
        parser.add_argument("-g", "--graph", action="store_true",
                            help="save graph info")
        opts = parser.parse_args()

        if opts.config is None:
            parser.error("Configuration file required.")
        if opts.start_time:
            start_time = opts.start_time
        else:
            start_time = datetime.utcnow()

        # [coords, station, area, scores], forward, start, pattern
        station_list, forward, start, pattern = read_config(opts.config)

        pattern_args = {
            "output_dir": opts.output_dir,
            "date": start_time.strftime("%Y%m%d"),
            "time": start_time.strftime("%H%M%S")
        }
        dir_output = build_filename("dir_output", pattern, pattern_args)
        if not os.path.exists(dir_output):
            sys.exit(1)
        ph = open(os.path.join(dir_output, "opts.pkl"), "rb")
        opts = pickle.load(ph)
        ph.close()

        graph = {}
        allpasses = {}
        for coords, station, area, scores in station_list:
            pattern_args["station"] = station
            graph[station] = Graph()
            graph[station].load(build_filename("file_graph", pattern, pattern_args) + ".npz")

#             print "---",station,"---"
#             print_matrix(graph[station].adj_matrix, ly=5)
#             print_matrix(graph[station].weight_matrix, ly=5, lx=-1)

#             allpasses[station] = get_passes_from_xml_file(os.path.join(opts.report,
#                                                           "acquisition-schedule-report." + station + ".xml"))
#             print len(allpasses[station]),allpasses[station]

#             for v in graph[station].neighbours(1):
#                 print v, " : ", allpasses[station][v].risetime, "->", graph[station].weight(1, v)

            ph = open(
                os.path.join(build_filename("dir_output", pattern, pattern_args), "allpasses.%s.pkl" % station), "rb")
            allpasses[station] = pickle.load(ph)
            ph.close()

        from trollsched.schedule import conflicting_passes
        totpas = []

        for s, sp in allpasses.items():
            logger.debug(f"s, len(sp): {s} {len(sp)}")
            totpas.extend(list(sp))
        passes = sorted(totpas, key=lambda x: x.risetime)
        cpg = conflicting_passes(passes, timedelta(seconds=600))
        logger.debug(f"ALLPASSES: {len(allpasses)}")
        logger.debug(f"PASSES: {len(passes)}")
        logger.debug(f"CONFLGRPS: {len(cpg)}")
        logger.debug(f"MAX: {max([len(g) for g in cpg])}")

        combined_stations(opts, pattern, station_list, graph, allpasses, start_time, start, forward)

    except Exception:
        logger.exception("Something wrong happened!")
        raise


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014-2024 PyTroll community

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The SwathBoundary class."""


import logging
import logging.handlers

import numpy as np
from pyorbital import geoloc, geoloc_instrument_definitions
from pyresample.boundary import Boundary

logger = logging.getLogger(__name__)

INSTRUMENT = {"avhrr/3": "avhrr",
              "avhrr/2": "avhrr",
              "avhrr-3": "avhrr",
              "mwhs-2": "mwhs2"}

class InstrumentNotSupported(Exception):
    """Exception to capture cases when instrument are (yet) not supported."""


class SwathBoundary(Boundary):
    """Boundaries for satellite overpasses."""

    def get_instrument_points(self, overpass, utctime,
                              scans_nb, scanpoints, scan_step=1):
        """Get the boundary points for a given overpass."""
        instrument, scan_angle = self.get_instrument_and_angle(overpass)

        sgeom = self.create_instrument_geometry(instrument, scans_nb, scanpoints, scan_step, scan_angle)

        times = sgeom.times(utctime)

        pixel_pos = geoloc.compute_pixels((self.orb.tle._line1,
                                           self.orb.tle._line2),
                                          sgeom, times)
        lons, lats, alts = geoloc.get_lonlatalt(pixel_pos, times)

        del alts
        return (lons.reshape(-1, len(scanpoints)),
                lats.reshape(-1, len(scanpoints)))

    def get_instrument_and_angle(self, overpass):
        """Get the instrument and angle for an overpass."""
        instrument = overpass.instrument
        if instrument == "modis":
            scan_angle = 55.0
            instrument = "avhrr"
        elif instrument == "viirs":
            scan_angle = 55.84
            instrument = "viirs"
        elif instrument == "iasi":
            scan_angle = 48.3
            instrument = "avhrr"
        elif overpass.satellite == "noaa 16":
            scan_angle = 55.25
            instrument = "avhrr"
        elif instrument.startswith("mersi"):
            scan_angle = 55.4
            instrument = "avhrr"
        elif instrument.startswith("slstr"):
            raise InstrumentNotSupported("SLSTR is a conical scanner, and currently not supported!")
        elif overpass.satellite.name.startswith("aws"):
            scan_angle = 55.25
            instrument = "avhrr"
        else:
            scan_angle = 55.25
        return instrument, scan_angle

    def create_instrument_geometry(self, instrument, scans_nb, scanpoints, scan_step, scan_angle):
        """Create an instrument geometry object."""
        instrument_fun = getattr(geoloc_instrument_definitions,
                                 INSTRUMENT.get(instrument, instrument))

        if instrument.startswith("avhrr"):
            sgeom = instrument_fun(scans_nb, scanpoints, scan_angle=scan_angle, frequency=100)
        elif instrument in ["ascat", ]:
            sgeom = instrument_fun(scans_nb, scanpoints)
        elif instrument in ["amsua", "mhs"]:
            sgeom = instrument_fun(scans_nb, scanpoints)
        elif instrument in ["mwhs2", ]:
            sgeom = instrument_fun(scans_nb, scanpoints)
        elif instrument in ["olci", ]:
            sgeom = instrument_fun(scans_nb, scanpoints)
        elif instrument == "viirs":
            sgeom = instrument_fun(scans_nb, scanpoints, scan_step=scan_step)
        elif instrument in ["mhs", "atms", "mwhs-2"]:
            sgeom = instrument_fun(scans_nb, scanpoints)
        elif instrument.startswith("slstr"):
            sgeom = instrument_fun(scans_nb, scanpoints)
        else:
            logger.warning("Instrument not tested: %s", instrument)
            sgeom = instrument_fun(scans_nb)
        return sgeom

    def __init__(self, overpass, scan_step=50, frequency=200):
        """Initialize the boundary.

        Arguments:
            overpass: the overpass to use
            scan_step: how many scans we should skip for a smaller boundary
            frequency: how much to decimate the top and bottom rows of the boundary.
        """
        # compute area covered by pass
        super().__init__()

        self.overpass = overpass
        self.orb = overpass.orb

        # compute sides

        scanlength_seconds = ((overpass.falltime - overpass.risetime).seconds +
                              (overpass.falltime - overpass.risetime).microseconds / 1000000.0)

        logger.debug("Instrument = %s", self.overpass.instrument)
        scan_step, sec_scan_duration, along_scan_reduce_factor = self.get_steps_and_duration(scan_step)

        # From pass length in seconds and the seconds for one scan derive the number of scans in the swath:
        scans_nb = scanlength_seconds / sec_scan_duration * along_scan_reduce_factor
        # Devide by the scan step to a reduced number of scans:
        scans_nb = np.floor(scans_nb / scan_step)
        scans_nb = int(max(scans_nb, 1))

        sides_lons, sides_lats = self.get_instrument_points(self.overpass,
                                                            overpass.risetime,
                                                            scans_nb,
                                                            np.array([0, self.overpass.number_of_fovs - 1]),
                                                            scan_step=scan_step)

        side_shape = sides_lons[::-1, 0].shape[0]
        nmod = 1

        if side_shape != scans_nb:
            nmod = side_shape // scans_nb
            logger.debug("Number of scan lines (%d) does not match number of scans (%d)",
                         side_shape, scans_nb)
            logger.info("Take every %d th element on the sides...", nmod)

        self.left_lons = sides_lons[::-1, 0][::nmod]
        self.left_lats = sides_lats[::-1, 0][::nmod]
        self.right_lons = sides_lons[:, 1][::nmod]
        self.right_lats = sides_lats[:, 1][::nmod]

        # compute bottom
        maxval = self.overpass.number_of_fovs
        rest = maxval % frequency
        mid_range = np.arange(rest / 2, maxval, frequency)
        if mid_range[0] == 0:
            start_idx = 1
        else:
            start_idx = 0

        reduced = np.hstack([0, mid_range[start_idx::], maxval - 1]).astype("int")

        lons, lats = self.get_instrument_points(self.overpass,
                                                overpass.falltime,
                                                1,
                                                reduced)

        self.bottom_lons = lons[0][::-1]
        self.bottom_lats = lats[0][::-1]

        # compute top
        lons, lats = self.get_instrument_points(self.overpass,
                                                overpass.risetime,
                                                1,
                                                reduced)

        self.top_lons = lons[0]
        self.top_lats = lats[0]

        return

    def get_steps_and_duration(self, scan_step):
        """Get the steps and duration for the instrument."""
        if self.overpass.instrument == "viirs":
            sec_scan_duration = 1.779166667
            along_scan_reduce_factor = 1
        elif self.overpass.instrument.startswith("avhrr"):
            sec_scan_duration = 1. / 6.
            along_scan_reduce_factor = 0.1
        elif self.overpass.instrument == "ascat":
            sec_scan_duration = 3.74747474747
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 1
        elif self.overpass.instrument == "amsua":
            sec_scan_duration = 8.
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 1
        elif self.overpass.instrument == "mhs":
            sec_scan_duration = 8./3.
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 1
        elif self.overpass.instrument == "mwhs2":
            sec_scan_duration = 8./3.
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 1
        elif self.overpass.instrument == "olci":
            # 3 minutes of data is 4091 300meter lines:
            sec_scan_duration = 0.04399902224395014
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 100
        elif self.overpass.instrument == "atms":
            sec_scan_duration = 8/3.
            along_scan_reduce_factor = 1
            # Overwrite the scan step
            scan_step = 1

        else:
            # Assume AVHRR!
            logmsg = ("Instrument scan duration not known. Setting it to AVHRR. Instrument: ")
            logger.info(logmsg + "%s", str(self.overpass.instrument))
            sec_scan_duration = 1. / 6.
            along_scan_reduce_factor = 0.1
        return scan_step, sec_scan_duration, along_scan_reduce_factor

    def decimate(self, ratio):
        """Remove points from the boundary."""
        length = len(self.top_lons)
        start = (length % ratio) / 2
        points = np.concatenate(([0], np.arange(start, length, ratio), [length - 1]))

        self.top_lons = self.top_lons[points]
        self.top_lats = self.top_lats[points]
        self.bottom_lons = self.bottom_lons[points]
        self.bottom_lats = self.bottom_lats[points]

        length = len(self.right_lons)
        start = (length % ratio) / 2
        points = np.concatenate(([0], np.arange(start, length, ratio), [length - 1]))

        self.right_lons = self.right_lons[points]
        self.right_lats = self.right_lats[points]
        self.left_lons = self.left_lons[points]
        self.left_lats = self.left_lats[points]

        return

    def contour(self):
        """Get the contour lon/lats."""
        lons = np.concatenate((self.top_lons,
                               self.right_lons[1:-1],
                               self.bottom_lons,
                               self.left_lons[1:-1]))
        lats = np.concatenate((self.top_lats,
                               self.right_lats[1:-1],
                               self.bottom_lats,
                               self.left_lats[1:-1]))
        return lons, lats
